with Ada.Characters.Latin_1;
with Ada.Direct_IO;
with Ada.Directories;
with Ada.Numerics.Elementary_Functions;
with Ada.Text_IO;
with AUnit.Assertions;
with Vectors;

use AUnit.Assertions;
use Vectors;

package body Vector_Tests is
    
    procedure Register_Tests(T: in out Test) is
        use AUnit.Test_Cases.Registration;
    begin
        Register_Routine(T, Test_Add'Access, "Test_Add");
        Register_Routine(T, Test_Are_Equal'Access, "Test_Are_Equal");
        Register_Routine(T, Test_Are_Not_Equal'Access, "Test_Are_Not_Equal");
        Register_Routine(T, Test_Are_Orthogonal'Access, "Test_Are_Orthogonal");
        Register_Routine(T, Test_Are_Not_Orthogonal'Access, 
            "Test_Are_Not_Orthogonal");
        Register_Routine(T, Test_Cross_Product'Access, "Test_Cross_Product");
        Register_Routine(T, Test_Distance'Access, "Test_Distance");
        Register_Routine(T, Test_Distance_To_Origin'Access, 
            "Test_Distance_To_Origin");
        Register_Routine(T, Test_Multiply'Access, "Test_Multiply");
        Register_Routine(T, Test_Put'Access, "Test_Put");
        Register_Routine(T, Test_Subtract'Access, "Test_Subtract");
        Register_Routine(T, Test_Scalar'Access, "Test_Scalar");
    end Register_Tests;
    
    -- ---------------------------------------------------------------

    function Name(T: Test) return Test_String is
        pragma Unreferenced(T);
    begin
        return Format("Vector_Tests");
    end Name;
    
    -- ---------------------------------------------------------------

    procedure Test_Add(T: in out Test_Cases.Test_Case'Class) is
        pragma Unreferenced(T);
        L: constant Vector := (2.0, 3.0, 4.0);
        R: constant Vector := (5.0, 6.0, 7.0);
        Actual: constant Vector := L + R;
        Expected: constant Vector := (L.X + R.X, L.Y + R.Y, L.Z + R.Z);
    begin
        Assert(Actual = Expected, "Test_Add failed.");
    end Test_Add;
    
    -- ---------------------------------------------------------------

    procedure Test_Are_Equal(T: in out Test_Cases.Test_Case'Class) is
        pragma Unreferenced(T);
        L: constant Vector := (2.0, 3.0, 4.0);
        R: constant Vector := (2.0, 3.0, 4.0);
    begin
        Assert(L = R, "Test_Are_Equal failed.");
    end Test_Are_Equal;
    
    -- ---------------------------------------------------------------

    procedure Test_Are_Not_Equal(T: in out Test_Cases.Test_Case'Class) is
        pragma Unreferenced(T);
        V: constant Vector := (2.0, 3.0, 4.0);
        X_False: constant Vector := (7.0, 3.0, 4.0);
        Y_False: constant Vector := (2.0, -3.0, 4.0);
        Z_False: constant Vector := (2.0, 3.0, -4.0);
    begin
        Assert(V /= X_False, "Test_Not_Are_Equal failed.");
        Assert(V /= Y_False, "Test_Not_Are_Equal failed.");
        Assert(V /= Z_False, "Test_Not_Are_Equal failed.");
    end Test_Are_Not_Equal;
    
    -- ---------------------------------------------------------------

    procedure Test_Are_Orthogonal(T: in out Test_Cases.Test_Case'Class) is
        pragma Unreferenced(T);
        L: constant Vector := (2.0, 3.0, -4.0);
        R: constant Vector := (5.0, 6.0, 7.0);
        Actual: constant Boolean := Are_Orthogonal(L, R);
    begin
        Assert(Actual, "Test_Are_Orthogonal failed.");
    end Test_Are_Orthogonal;
    
    -- ---------------------------------------------------------------

    procedure Test_Are_Not_Orthogonal(T: in out Test_Cases.Test_Case'Class) is
        pragma Unreferenced(T);
        L: constant Vector := (2.0, 3.0, 4.0);
        R: constant Vector := (5.0, 6.0, 7.0);
        Actual: constant Boolean := Are_Orthogonal(L, R);
    begin
        Assert(not Actual, "Test_Are_Not_Orthogonal failed.");
    end Test_Are_Not_Orthogonal;
    
    -- ---------------------------------------------------------------

    procedure Test_Cross_Product(T: in out Test_Cases.Test_Case'Class) is
        pragma Unreferenced(T);
        L: constant Vector := (2.0, 3.0, 4.0);
        R: constant Vector := (5.0, 6.0, 7.0);
        Actual: constant Vector := Cross_Product(L, R);
        Expected: constant Vector := (L.Y * R.Z - L.Z * R.Y, 
            L.Z * R.X - L.X * R.Z, 
            L.X * R.Y - L.Y * R.X);
    begin
        Assert(Expected = Actual, "Test_Cross_Product failed.");
    end Test_Cross_Product;
    
    -- ---------------------------------------------------------------

    procedure Test_Distance(T: in out Test_Cases.Test_Case'Class) is
        pragma Unreferenced(T);
        use Ada.Numerics.Elementary_Functions;
        
        L: constant Vector := (2.0, 3.0, 4.0);
        R: constant Vector := (5.0, 6.0, 7.0);
        Actual: constant Float := Distance(L, R);
        Expected: constant Float := Sqrt(
            (L.X - R.X)**2 + (L.Y - R.Y)**2 + (L.Z - R.Z)**2
        );
    begin
        Assert(Expected = Actual, "Test_Distance failed.");
    end Test_Distance;
    
    -- ---------------------------------------------------------------

    procedure Test_Distance_To_Origin(T: in out Test_Cases.Test_Case'Class) is
        pragma Unreferenced(T);
        use Ada.Numerics.Elementary_Functions;
        
        Item: constant Vector := (2.0, 3.0, 4.0);
        Actual: constant Float := Distance_To_Origin(Item);
        Expected: constant Float := Sqrt(Item.X**2 + Item.Y**2 + Item.Z**2);
    begin
        Assert(Expected = Actual, "Test_Distance_To_Origin failed.");
    end Test_Distance_To_Origin;
    
    -- ---------------------------------------------------------------

    procedure Test_Multiply(T: in out Test_Case'Class) is
        pragma Unreferenced(T);
        L: constant Vector := (2.0, 3.0, 4.0);
        R: constant Float := 5.0;
        Actual: constant Vector := L * R;
        Expected: constant Vector := (L.X * R, L.Y * R, L.Z * R);
    begin
        Assert(Actual = Expected, "Test_Multiply failed.");
    end Test_Multiply;

    -- ---------------------------------------------------------------

    procedure Test_Put(T: in out Test_Cases.Test_Case'Class) is
        pragma Unreferenced(T);
        use Ada.Text_IO;
        
        Stdout: constant File_Type := Standard_Output;
        Put_File_Name: constant String := "vector_put_test.txt";
        Put_File: File_Type;
        
        L: constant Vector := (2.0, 3.0, 4.55);
        Expected: constant String := "(2.00000, 3.00000, 4.55000)"
            & Ada.Characters.Latin_1.LF;
    begin
        -- Create the output file and redirect output
        Create(Put_File, Out_File, Put_File_Name); 
        Set_Output(Put_File);
        
        -- Invoke Put
        Vectors.Put(L);
        
        -- Redirect output to stdout and close the file
        Close(Put_File);
        Set_Output(Stdout);
        
        -- Read file
        declare
            File_Size: constant Natural := 
                Natural(Ada.Directories.Size(Put_File_Name));
            Actual: String(1 .. File_Size);
        begin
            Actual := Read_File(Put_File_Name, File_Size);
            Assert(Expected = Actual, "Test_Put failed.");
        end;
    end Test_Put;
    
    -- ---------------------------------------------------------------

    function Read_File(File_Name: String; File_Size: Natural) return String is
        subtype File_String is String(1 .. File_Size);
        package File_String_IO is new Ada.Direct_IO(File_String);
        
        File: File_String_IO.File_Type;
        Contents: File_String;
    begin
        File_String_IO.Open(File, File_String_IO.In_File, File_Name);
        File_String_IO.Read(File, Contents);
        File_String_IO.Close(File);
        
        return Contents;
    end Read_File;
    
    -- ---------------------------------------------------------------

    procedure Test_Subtract(T: in out Test_Cases.Test_Case'Class) is
        pragma Unreferenced(T);
        L: constant Vector := (2.0, 3.0, 4.0);
        R: constant Vector := (5.0, 6.0, 7.0);
        Actual: constant Vector := L - R;
        Expected: constant Vector := (L.X - R.X, L.Y - R.Y, L.Z - R.Z);
    begin
        Assert(Actual = Expected, "Test_Subtract failed.");
    end Test_Subtract;
    
    -- ---------------------------------------------------------------

    procedure Test_Scalar(T: in out Test_Cases.Test_Case'Class) is
        pragma Unreferenced(T);
        L: constant Vector := (2.0, 3.0, 4.0);
        R: constant Vector := (5.0, 6.0, 7.0);
        Actual: constant Float := L * R;
        Expected: constant Float := L.X * R.X + L.Y * R.Y + L.Z * R.Z;
    begin
        Assert(Expected = Actual, "Test_Scalar failed.");
    end Test_Scalar;
    
end Vector_Tests;
