with Ada.Integer_Text_IO;
with AUnit.Assertions;
with Binary_Trees;

use AUnit.Assertions;

package body Binary_Trees_Tests is
    
    procedure Put_Item(Item: Integer) is
        use Ada.Integer_Text_IO;
    begin
        Put(Item);
    end Put_Item;

    -- ---------------------------------------------------------------

    package Integer_Trees is new Binary_Trees(
        Item_Type => Integer, 
        "<"       => "<", 
        "="       => "=", 
        Put_Item  => Put_Item
    );
    use Integer_Trees;

    -- ---------------------------------------------------------------

    procedure Create_2Level_Test_Tree(Tree: in out Binary_Tree_Access) is
    begin
        Tree := Create(5);
        Add_Item(Tree, 3);
        Add_Item(Tree, 7);
    end Create_2Level_Test_Tree;

    -- ---------------------------------------------------------------

    procedure Create_3Level_Test_Tree(Tree: in out Binary_Tree_Access) is
    begin
        Tree := Create(5);
        Add_Item(Tree, 3);
        Add_Item(Tree, 7);
        Add_Item(Tree, 1);
        Add_Item(Tree, 6);
        Add_Item(Tree, 8);
    end Create_3Level_Test_Tree;

    -- ---------------------------------------------------------------

    procedure Test_Add_Item(T: in out Test_Case'Class) is
        pragma Unreferenced(T);
        Tree: Binary_Tree_Access := Create(1);
    begin
        Add_Item(Tree, 2);
        Assert(Has_Item(Tree, 2), "Test_Add_Item failed");
    end Test_Add_Item;

    -- ---------------------------------------------------------------

    procedure Test_Add_Items_In_Sequence(T: in out Test_Case'Class) is
        pragma Unreferenced(T);
        Tree: Binary_Tree_Access := Create(1);
        Num_Items: constant Natural := 10;
    begin
        for I in 2 .. Num_Items loop
            Add_Item(Tree, I);
        end loop;

        for I in 1 .. Num_Items loop
            Assert(Has_Item(Tree, I), "Test_Add_Items_In_Sequence failed");
        end loop;
    end Test_Add_Items_In_Sequence;

    -- ---------------------------------------------------------------

    procedure Test_Add_Items_In_Reverse_Sequence(T: in out Test_Case'Class) is
        pragma Unreferenced(T);
        Tree: Binary_Tree_Access := Create(1);
        Num_Items: constant Natural := 10;
    begin
        for I in reverse 2 .. Num_Items loop
            Add_Item(Tree, I);
        end loop;

        for I in 1 .. Num_Items loop
            Assert(Has_Item(Tree, I), 
                "Test_Add_Items_In_Reverse_Sequence failed");
        end loop;
    end Test_Add_Items_In_Reverse_Sequence;

    -- ---------------------------------------------------------------

    procedure Test_Node(Node: Binary_Tree_Access; 
                        Expected_Item: Integer; 
                        Has_Left: Boolean; 
                        Has_Right: Boolean; 
                        Expected_Left: Integer  := 0; 
                        Expected_Right: Integer := 0) is
    begin
        Assert(Node /= null, "Node must not be zero");
        Assert(Node.all.Item = Expected_Item, "Node item was incorrect");

        if Has_Left then
            Assert(Node.all.Left /= null, 
                "Node's left child must not be null");
            Assert(Node.all.Left.all.Item = Expected_Left,
                "Node's left item was incorrect");
        end if;

        if Has_Right then
            Assert(Node.all.Right /= null, 
                "Node's right child must not be null");
            Assert(Node.all.Right.all.Item = Expected_Right,
                "Node's right item was incorrect");
        end if;
    end Test_Node;

    -- ---------------------------------------------------------------

    procedure Test_Add_Items_In_Mixed_Order(T: in out Test_Case'Class) is
        pragma Unreferenced(T);
        Items: constant array (1..6) of Natural := (5, 4, 2, 1, 6, 3);
        Tree: Binary_Tree_Access := Create(Items(Items'First));
    begin
        for I in Items'First+1 .. Items'Last loop
            Add_Item(Tree, Items(I));
        end loop;

        Test_Node(Tree,                 5, True,  True,  4, 6);
        Test_Node(Tree.Right,           6, False, False);
        Test_Node(Tree.Left,            4, True,  False, 2);
        Test_Node(Tree.Left.Left,       2, True,  True,  1, 3);
        Test_Node(Tree.Left.Left.Left,  1, False, False);
        Test_Node(Tree.Left.Left.Right, 3, False, False);
    end Test_Add_Items_In_Mixed_Order;

    -- ---------------------------------------------------------------

    procedure Test_Add_Item_Fails_For_Duplicate(T: in out Test_Case'Class) is
        pragma Unreferenced(T);
        Tree: Binary_Tree_Access := Create(1);
    begin
        Add_Item(Tree, 1);
        Assert(False, 
            "Test_Add_Item_Fails_For_Duplicate should raise exception");
    exception
        when Item_Already_In_Tree_Exception => return;
        when others => Assert(False, 
            "Test_Add_Item_Fails_For_Duplicate raised incorrect exception");
    end Test_Add_Item_Fails_For_Duplicate;

    -- ---------------------------------------------------------------

    procedure Test_Add_Item_Fails_For_Leaf(T: in out Test_Case'Class) is
        pragma Unreferenced(T);
        Tree: Binary_Tree_Access := Create(1);
    begin
        Add_Item(Tree, 2);
        Add_Item(Tree, 3);
        Add_Item(Tree, 2);
        Assert(False, "Test_Add_Item_Fails_For_Leaf should raise exception");
    exception
        when Item_Already_In_Tree_Exception => return;
        when others => Assert(False, 
            "Test_Add_Item_Fails_For_Leaf raised incorrect exception");
    end Test_Add_Item_Fails_For_Leaf;

    -- ---------------------------------------------------------------

    procedure Test_Create(T: in out Test_Case'Class) is
        pragma Unreferenced(T);
        Tree: constant Binary_Tree_Access := Create(5);
    begin
        Assert(Tree.all.Item = 5, "Create should set item incorrectly");
        Assert(Tree.all.Parent = null, "Create should set Parent to null");
        Assert(Tree.all.Left = null, "Create should set Left to null");
        Assert(Tree.all.Right = null, "Create should set Right to null");
    end Test_Create;

    -- ---------------------------------------------------------------

    procedure Test_Has_Left_Child_Only(T: in out Test_Case'Class) is
        pragma Unreferenced(T);
        Tree: Binary_Tree_Access := Create(3);
    begin
        Add_Item(Tree, 2);
        Assert(Has_Children(Tree), "Test_Has_Left_Child_Only failed");
    end Test_Has_Left_Child_Only;

    -- ---------------------------------------------------------------

    procedure Test_Has_Right_Child_Only(T: in out Test_Case'Class) is
        pragma Unreferenced(T);
        Tree: Binary_Tree_Access := Create(3);
    begin
        Add_Item(Tree, 4);
        Assert(Has_Children(Tree), "Test_Has_Right_Child_Only failed");
    end Test_Has_Right_Child_Only;

    -- ---------------------------------------------------------------

    procedure Test_Has_Multiple_Children(T: in out Test_Case'Class) is
        pragma Unreferenced(T);
        Tree: Binary_Tree_Access := Create(3);
    begin
        Add_Item(Tree, 4);
        Add_Item(Tree, 2);
        Assert(Has_Children(Tree), "Test_Has_Multiple_Children failed");
    end Test_Has_Multiple_Children;

    -- ---------------------------------------------------------------

    procedure Test_Has_No_Children(T: in out Test_Case'Class) is
        pragma Unreferenced(T);
        Tree: constant Binary_Tree_Access := Create(3);
    begin
        Assert(not Has_Children(Tree), "Test_Has_No_Children failed");
    end Test_Has_No_Children;

    -- ---------------------------------------------------------------

    procedure Test_Has_Item(T: in out Test_Case'Class) is
        pragma Unreferenced(T);
        Tree: Binary_Tree_Access := Create(1);
    begin
        for I in 2..10 loop
            Add_Item(Tree, I);
            Assert(Has_Item(Tree, I), "Test_Has_Item failed");
        end loop;

        for I in 1..10 loop
            Assert(Has_Item(Tree, I), "Test_Has_Item failed");
        end loop;
    end Test_Has_Item;

    -- ---------------------------------------------------------------

    procedure Test_Has_Item_Reverse(T: in out Test_Case'Class) is
        pragma Unreferenced(T);
        Tree: Binary_Tree_Access := Create(1);
    begin
        for I in reverse 2..10 loop
            Add_Item(Tree, I);
            Assert(Has_Item(Tree, I), "Test_Has_Item_Reverse failed");
        end loop;

        for I in 1..10 loop
            Assert(Has_Item(Tree, I), "Test_Has_Item_Reverse failed");
        end loop;
    end Test_Has_Item_Reverse;

    -- ---------------------------------------------------------------

    procedure Test_Has_Item_Should_Be_False(T: in out Test_Case'Class) is
        pragma Unreferenced(T);
        Tree: constant Binary_Tree_Access := Create(1);
    begin
        for I in 2..10 loop
            Assert(not Has_Item(Tree, I), 
                "Test_Has_Item_Should_Be_False failed");
        end loop;
    end Test_Has_Item_Should_Be_False;

    -- ---------------------------------------------------------------

    procedure Test_Get_Height_Single(T: in out Test_Case'Class) is
        pragma Unreferenced(T);
        Tree: constant Binary_Tree_Access := Create(1);
    begin
        Assert(Get_Height(Tree) = 1, "Test_Get_Height_Single failed");
    end Test_Get_Height_Single;

    -- ---------------------------------------------------------------

    procedure Test_Get_Height_3_Levels(T: in out Test_Case'Class) is
        pragma Unreferenced(T);
        Tree: Binary_Tree_Access := Create(5);
    begin
        Assert(Get_Height(Tree) = 1, "Test_Get_Height_3_Levels failed");
        Add_Item(Tree, 3);
        Assert(Get_Height(Tree) = 2, "Test_Get_Height_3_Levels failed");
        Add_Item(Tree, 7);
        Assert(Get_Height(Tree) = 2, "Test_Get_Height_3_Levels failed");
        Add_Item(Tree, 1);
        Assert(Get_Height(Tree) = 3, "Test_Get_Height_3_Levels failed");
        Add_Item(Tree, 6);
        Assert(Get_Height(Tree) = 3, "Test_Get_Height_3_Levels failed");
        Add_Item(Tree, 8);
        Assert(Get_Height(Tree) = 3, "Test_Get_Height_3_Levels failed");
    end Test_Get_Height_3_Levels;

    -- ---------------------------------------------------------------

    procedure Test_Get_Height_Empty(T: in out Test_Case'Class) is
        pragma Unreferenced(T);
        Tree: constant Binary_Tree_Access := Create(1);
    begin
        Assert(Get_Height(Tree) = 1, "Test_Get_Height_Empty failed");
    end Test_Get_Height_Empty;

    -- ---------------------------------------------------------------

    procedure Test_Get_Height_Null(T: in out Test_Case'Class) is
        pragma Unreferenced(T);
    begin
        Assert(Get_Height(null) = 0, "Test_Get_Height_Null failed");
    end Test_Get_Height_Null;

    -- ---------------------------------------------------------------

    procedure Test_Remove_Item(T: in out Test_Case'Class) is
        pragma Unreferenced(T);
        Tree: Binary_Tree_Access;
    begin
        Create_2Level_Test_Tree(Tree);
        Remove_Item(Tree, 3);

        Assert(not Has_Item(Tree, 3), "Test_Remove_Item did not remove item");
        Assert(Has_Item(Tree, 5), "Test_Remove_Item removed other item");
        Assert(Has_Item(Tree, 7), "Test_Remove_Item removed other item");

        Remove_All(Tree);
        Create_2Level_Test_Tree(Tree);

        Remove_Item(Tree, 5);

        Assert(Has_Item(Tree, 3), "Test_Remove_Item removed other item");
        Assert(not Has_Item(Tree, 5), 
            "Test_Remove_Item did not remove item");
        Assert(Has_Item(Tree, 7), "Test_Remove_Item removed other item");

        Create_2Level_Test_Tree(Tree);
        Remove_Item(Tree, 7);

        Assert(Has_Item(Tree, 3), "Test_Remove_Item removed other item");
        Assert(Has_Item(Tree, 5), "Test_Remove_Item removed other item");
        Assert(not Has_Item(Tree, 7), "Test_Remove_Item did not remove item");
    end Test_Remove_Item;

    -- ---------------------------------------------------------------

    procedure Test_Remove_Subtree_Root(T: in out Test_Case'Class) is
        pragma Unreferenced(T);
        Tree: Binary_Tree_Access;
    begin
        Create_3Level_Test_Tree(Tree);
        Remove_Item(Tree, 7);

        Assert(Has_Item(Tree, 1), "Test_Remove_Item removed other item");
        Assert(Has_Item(Tree, 3), "Test_Remove_Item removed remove item");
        Assert(Has_Item(Tree, 5), "Test_Remove_Item removed other item");
        Assert(Has_Item(Tree, 6), "Test_Remove_Item removed subtree");
        Assert(not Has_Item(Tree, 7), "Test_Remove_Item did not remove item");
        Assert(Has_Item(Tree, 8), "Test_Remove_Item removed subtree");
    end Test_Remove_Subtree_Root;

    -- ---------------------------------------------------------------

    procedure Test_Remove_Item_Should_Fail(T: in out Test_Case'Class) is
        pragma Unreferenced(T);
        Tree: Binary_Tree_Access;
    begin
        Create_2Level_Test_Tree(Tree);
        Remove_Item(Tree, 2);
        Assert(False, "Remove_Item should raise exception");
    exception
        when Item_Not_Found_Exception => return;
        when others => Assert(False, "Wrong exception");
    end Test_Remove_Item_Should_Fail;

    -- ---------------------------------------------------------------

    procedure Test_Remove_All(T: in out Test_Case'Class) is
        pragma Unreferenced(T);
        Tree: Binary_Tree_Access;
    begin
        Create_3Level_Test_Tree(Tree);
        Remove_All(Tree);
        Assert(Tree = null, "Remove all should set tree to null");
    end Test_Remove_All;

    -- ---------------------------------------------------------------

    procedure Test_Remove_All_Empty(T: in out Test_Case'Class) is
        pragma Unreferenced(T);
        Tree: Binary_Tree_Access := Create(5);
    begin
        Remove_All(Tree);
        Assert(Tree = null, "Remove all should set tree to null");
    end Test_Remove_All_Empty;

    -- ---------------------------------------------------------------

    procedure Test_Remove_All_Null(T: in out Test_Case'Class) is
        pragma Unreferenced(T);
        Tree: Binary_Tree_Access;
    begin
        Remove_All(Tree);
        Assert(Tree = null, "Remove all should set tree to null");
    end Test_Remove_All_Null;

    -- ---------------------------------------------------------------
    
    procedure Register_Tests(T: in out Test) is
        use AUnit.Test_Cases.Registration;
    begin
        Register_Routine(T, Test_Add_Item'Access, "Test_Add_Item");
        Register_Routine(T, Test_Add_Items_In_Sequence'Access, 
            "Test_Add_Items_In_Sequence");
        Register_Routine(T, Test_Add_Items_In_Reverse_Sequence'Access, 
            "Test_Add_Items_In_Reverse_Sequence");
        Register_Routine(T, Test_Add_Items_In_Mixed_Order'Access, 
            "Test_Add_Items_In_Mixed_Order");
        Register_Routine(T, Test_Add_Item_Fails_For_Duplicate'Access, 
            "Test_Add_Item_Fails_For_Duplicate");
        Register_Routine(T, Test_Add_Item_Fails_For_Leaf'Access, 
            "Test_Add_Item_Fails_For_Leaf");
        Register_Routine(T, Test_Create'Access, "Test_Create");
        Register_Routine(T, Test_Has_Left_Child_Only'Access, 
            "Test_Has_Left_Child_Only");
        Register_Routine(T, Test_Has_Right_Child_Only'Access, 
            "Test_Has_Right_Child_Only");
        Register_Routine(T, Test_Has_Multiple_Children'Access, 
            "Test_Has_Multiple_Children");
        Register_Routine(T, Test_Has_No_Children'Access, 
            "Test_Has_No_Children");
        Register_Routine(T, Test_Has_Item'Access, "Test_Has_Item");
        Register_Routine(T, Test_Has_Item_Reverse'Access, 
            "Test_Has_Item_Reverse");
        Register_Routine(T, Test_Has_Item_Should_Be_False'Access, 
            "Test_Has_Item_Should_Be_False");
        Register_Routine(T, Test_Get_Height_3_Levels'Access, 
            "Test_Get_Height_3_Levels");
        Register_Routine(T, Test_Get_Height_Single'Access, 
            "Test_Get_Height_Single");
        Register_Routine(T, Test_Get_Height_Empty'Access, 
            "Test_Get_Height_Empty");
        Register_Routine(T, Test_Get_Height_Null'Access, 
            "Test_Get_Height_Null");
        Register_Routine(T, Test_Remove_Item'Access, "Test_Remove_Item");
        Register_Routine(T, Test_Remove_Subtree_Root'Access, 
            "Test_Remove_Subtree_Root");
        Register_Routine(T, Test_Remove_Item_Should_Fail'Access, 
            "Test_Remove_Item_Should_Fail");
        Register_Routine(T, Test_Remove_All'Access, "Test_Remove_All");
        Register_Routine(T, Test_Remove_All_Empty'Access, 
            "Test_Remove_All_Empty");
        Register_Routine(T, Test_Remove_All_Null'Access, 
            "Test_Remove_All_Null");
    end Register_Tests;
    
    -- ---------------------------------------------------------------

    function Name(T: Test) return Test_String is
        pragma Unreferenced(T);
    begin
        return Format("Binary_Trees_Tests");
    end Name;
    
end Binary_Trees_Tests;
